///////////////////////////////////////////////////////////////////////////////
// File: Service.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "Service.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/Win32Tools/Win32Exception.h"
#include "JetByteTools/Win32Tools/SEHException.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"

#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/AddressIPv4.h"
#include "JetByteTools/SocketTools/IAddressType.h"
#include "JetByteTools/SocketTools/ConnectionLimiter.h"
#include "JetByteTools/SocketTools/Socket.h"
#include "JetByteTools/SocketTools/UsesWinsock.h"

#include "JetByteTools/ServiceTools/INotifyProgress.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::GetCurrentDirectory;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CStringConverter;

using JetByteTools::Win32::CWin32Exception;
using JetByteTools::Win32::CSEHException;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;

using JetByteTools::Socket::CSequencedStreamSocketAllocator;
using JetByteTools::Socket::CAddressIPv4;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::CSocket;
using JetByteTools::Socket::CUsesWinsock;

using JetByteTools::Service::INotifyProgress;
using JetByteTools::Service::IShutdownService;
using JetByteTools::Service::IManageServiceNotifications;
using JetByteTools::Service::ServiceTypes;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static unsigned short GetPort(
   unsigned short port);

///////////////////////////////////////////////////////////////////////////////
// CService
///////////////////////////////////////////////////////////////////////////////

CService::CService()
   :  JetByteTools::Service::CService(
         *this,
         _T("EchoServerMultiInstanceService"),
         _T("JetByte Echo Server Multi Instance Service")),
      m_pPool(nullptr),
      m_pSocketAllocator(nullptr),
      m_pBufferAllocator(nullptr),
      m_pLimiter(nullptr),
      m_pServer(nullptr),
      m_isActive(false)
{

}

ServiceTypes::ExitCode CService::InitialiseService(
   IShutdownService & /*shutdown*/,
   INotifyProgress &notify,
   IManageServiceNotifications & /*notificationManager*/,
   const DWORD argc,
   const PTSTR *ppArgv)
{
   OutputEx(_T("Service: \"") + GetServiceName() + _T("\" initialising"));

   OutputEx(_T("Argument count: ") + ToString(argc));

   for (DWORD i = 0; i < argc; ++i)
   {
      OutputEx(_T("Argument: ") + ToString(i) + _T(" \"") + ppArgv[i] + _T("\""));
   }

   const _tstring logFileName = GetCurrentDirectory() + _T("\\EchoServerMultiInstanceService-") + GetInstanceName() + _T(".log");

   OutputEx(_T("Switching log file to: ") + logFileName);

   JetByteTools::Core::CDebugTrace::Instance().SetLogName(logFileName);

   OutputEx(_T("Service: \"") + GetServiceName() + _T("\" initialising"));

   notify.Notify(1000);

   m_pPool = new CIOPool(0);      // number of threads (0 = 2 x processors)

   m_pPool->Start();

   notify.Notify(1000);

   m_pSocketAllocator = new CSequencedStreamSocketAllocator(
      10,                           // max number of sockets to keep in the pool
      10,                           // max number of buffers to pool in the socket
      10);                          // max number of consecutive operations on this socket

   m_pBufferAllocator = new CBufferAllocator(
      1024,                         // buffer size
      10);                          // max number of buffers to keep in the pool

   notify.Notify(1000);

   const CAddressIPv4 address(
      INADDR_ANY,                   // address to listen on
      GetPort(5050));               // port to listen on

   const size_t listenBacklog = 100;

   m_pLimiter = new CConnectionLimiter(CConnectionLimiter::NoLimit);

   m_pServer = new CSocketServer(
      "Welcome to echo server\r\n",
      address,
      listenBacklog,
      *m_pPool,
      *m_pSocketAllocator,
      *m_pBufferAllocator,
      *m_pLimiter);

   notify.Notify(5000);

   m_pServer->Start();

   notify.Notify(1000);

   StartupComplete();

   OutputEx(_T("Service: \"") + GetServiceName() + _T("\" running"));

   return ServiceTypes::ExitSuccess;
}

void CService::ContinueService(
   INotifyProgress & /*notify*/)
{
   if (!m_isActive)
   {
      m_pServer->StartAcceptingConnections();

      m_isActive = true;
   }
}

void CService::PauseService(
   INotifyProgress & /*notify*/)
{
   if (m_isActive)
   {
      m_pServer->StopAcceptingConnections();

      m_isActive = false;
   }
}

void CService::StopService(
   INotifyProgress & /*notify*/,
   ServiceTypes::StopReason /*stopReason*/)
{
   if (m_isActive)
   {
      m_pServer->StopAcceptingConnections();

      m_isActive = false;
   }
}

ServiceTypes::ExitCode CService::UninitialiseService(
   INotifyProgress &notify)
{
   OutputEx(_T("Service: \"") + GetServiceName() + _T("\" shutting down"));

   if (m_pServer)
   {
      OutputEx(_T("Shutting down socket server."));

      notify.Notify(5000);

      if (!m_pServer->WaitForShutdownToComplete(5000))
      {
         OutputEx(_T("Socket server failed to shut down."));

         m_pServer->ForceShutdown();

         OutputEx(_T("Socket server forcibly shut down."));
      }
   }

   if (m_pPool)
   {
      OutputEx(_T("Shutting down I/O pool."));

      notify.Notify(5000);

      m_pPool->WaitForShutdownToComplete();
   }

   if (m_pBufferAllocator)
   {
      OutputEx(_T("Flushing buffer allocator."));

      notify.Notify(5000);

      if (m_pBufferAllocator->Flush())
      {
         OutputEx(_T("Buffers were still allocated when the allocator was flushed..."));
      }
   }

   if (m_pSocketAllocator)
   {
      OutputEx(_T("Flushing socket allocator."));

      notify.Notify(5000);

      if (m_pSocketAllocator->ReleaseSockets())
      {
         OutputEx(_T("Sockets were still allocated when the allocator was flushed..."));
      }
   }

   OutputEx(_T("Deleting objects..."));

   delete m_pServer;

   m_pServer = nullptr;

   delete m_pLimiter;

   m_pLimiter = nullptr;

   delete m_pPool;

   m_pPool = nullptr;

   delete m_pBufferAllocator;

   m_pBufferAllocator = nullptr;

   delete m_pSocketAllocator;

   m_pSocketAllocator = nullptr;

   OutputEx(_T("Service: \"") + GetServiceName() + _T("\" shut down"));

   return ServiceTypes::ExitSuccess;
}

ServiceTypes::ExitCode CService::OnException(
   const CException &e)
{
   OutputEx(_T("CService::OnException() - Exception: ") + e.GetDetails());

   return ServiceTypes::UnhandledException;
}

ServiceTypes::ExitCode CService::OnException(
   const CSEHException &e)
{
   OutputEx(_T("CService::OnException() - Exception: ") + e.GetDetails());

   return ServiceTypes::UnhandledException;
}

ServiceTypes::ExitCode CService::OnException(
   const std::exception &e)
{
   OutputEx(_T("CService::OnException() - Exception: ") + CStringConverter::AtoT(e.what()));

   return ServiceTypes::UnhandledException;
}

ServiceTypes::ExitCode CService::OnException()
{
   OutputEx(_T("CService::OnException() - Unexpected Exception"));

   return ServiceTypes::UnhandledException;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static unsigned short GetPort(
   unsigned short port)
{
   CUsesWinsock usesWinsock;

   bool done = false;

   while (!done)
   {
      const CAddressIPv4 address(
         INADDR_ANY,                   // address to listen on
         port);                        // port to listen on

      SOCKET s = ::WSASocket(address.Type().Family(), SOCK_STREAM, address.Type().Protocol(), nullptr, 0, WSA_FLAG_OVERLAPPED);

      if (s == INVALID_SOCKET)
      {
         throw CWin32Exception(_T("CSocket::CreateOutboundSocket()"), ::WSAGetLastError());
      }

      CSocket socket(s);

      if (NO_ERROR == socket.TryBind(address))
      {
         done = true;
      }
      else
      {
         port++;
      }
   }

   return port;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: Service.cpp
///////////////////////////////////////////////////////////////////////////////
